'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var path = require('node:path');
var uniCliShared = require('@dcloudio/uni-cli-shared');
var shared = require('@plugin-light/shared');
var tComm = require('t-comm');
var os = require('node:os');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var os__default = /*#__PURE__*/_interopDefaultLegacy(os);

const ENTRY = 'main.ts';
const APP_VUE = 'App.vue';
function slash(p) {
    return p.replace(/\\/g, '/');
}
const IS_WINDOWS = os__default["default"].platform() === 'win32';
const CHUNK_FILE_NAME_BLACK_LIST = ['main', 'pages.json', 'manifest.json'];
const MAIN = 'MAIN';
const cssLangs = '\\.(css|less|sass|scss|styl|stylus|pcss|postcss)($|\\?)';
const cssLangRE = new RegExp(cssLangs);
const directRequestRE = /(\?|&)direct\b/;
const uniPagePrefix = 'uniPage://';
const uniComponentPrefix = 'uniComponent://';
function virtualComponentPath(filepath) {
    return uniComponentPrefix + uniCliShared.encodeBase64Url(filepath);
}
function parseVirtualPagePath(uniPageUrl) {
    return uniCliShared.decodeBase64Url(uniPageUrl.replace(uniPagePrefix, ''));
}
function parseVirtualComponentPath(uniComponentUrl) {
    return uniCliShared.decodeBase64Url(uniComponentUrl.replace(uniComponentPrefix, ''));
}
function isUniPageUrl(id) {
    return id.startsWith(uniPagePrefix);
}
function isUniComponentUrl(id) {
    return id.startsWith(uniComponentPrefix);
}
const isCSSRequest = (request) => cssLangRE.test(request) && !directRequestRE.test(request);
const KNOWN_ASSET_TYPES = [
    // images
    'png',
    'jpe?g',
    'gif',
    'svg',
    'ico',
    'webp',
    'avif',
    // media
    'mp4',
    'webm',
    'ogg',
    'mp3',
    'wav',
    'flac',
    'aac',
    // fonts
    'woff2?',
    'eot',
    'ttf',
    'otf',
    // other
    'pdf',
    'txt',
];
const DEFAULT_ASSETS_RE = new RegExp(`\\.(${KNOWN_ASSET_TYPES.join('|')})(\\?.*)?$`);
const EXTNAME_JS_RE = /\.(js|jsx|ts|uts|tsx|mjs)$/;
function removeExt(str) {
    return str.split('?')[0].replace(/\.\w+$/g, '');
}
function isVueJs(id) {
    return id.includes('\0plugin-vue:export-helper');
}
function normalizePath(id) {
    return path__default["default"].posix.normalize(IS_WINDOWS ? slash(id) : id);
}
const PAGES_JSON_JS = '/pages-json-js';
const ANALYZE_BLACK_LIST = [
    'node_modules/vuex',
    'node_modules/@vue',
    'node_modules/@dcloudio',
    'node_modules/@dcloudio',
    '?vue&type=style',
    'pages-json-js',
    '/node_modules/lodash-es',
    '/node_modules/@babel',
];

const importerCache = new Map();
const circularSet = new Set();
function findImporters({ id, importChain, getModuleInfo, getModuleIds, pre = [], inputDir = '', iterativeMap = {}, }) {
    let parsedId = id;
    if (isUniPageUrl(parsedId)) {
        parsedId = normalizePath(path__default["default"].resolve(inputDir, parseVirtualPagePath(parsedId)));
    }
    if (isUniComponentUrl(parsedId)) {
        parsedId = normalizePath(path__default["default"].resolve(inputDir, parseVirtualComponentPath(parsedId)));
    }
    // 虚拟页面的引用者一定为空
    if (parsedId.endsWith('.vue') && uniCliShared.isMiniProgramPageFile(parsedId, process.env.UNI_INPUT_DIR)) {
        return [];
    }
    // 非虚拟组件的引用者只有虚拟组件，没必要分析
    if (parsedId === id && id.endsWith('.vue')) {
        return [];
    }
    const moduleInfo = getModuleInfo(id);
    iterativeMap[parsedId] = {
        parsedId: parsedId !== id ? parsedId : undefined,
        dynamic: moduleInfo?.dynamicImporters || [],
        importers: moduleInfo?.importers || [],
    };
    // pre 只存真正的组件/脚本
    if (pre.includes(parsedId)) {
        const res = [...importChain,
            // 出现循环依赖，就假装它被 main.ts 依赖，自然不会被处理
            path__default["default"].resolve(inputDir, ENTRY),
        ];
        circularSet.add(parsedId);
        importerCache.set(parsedId, res);
        return res;
    }
    if (importerCache.has(parsedId)) {
        return importerCache.get(parsedId) || [];
    }
    const allImporters = [
        ...(moduleInfo?.importers || []),
        ...(moduleInfo?.dynamicImporters || []),
    ];
    for (const importer of allImporters) {
        let parsedImporter = importer;
        let virtualImporter = importer;
        if (isUniPageUrl(importer)) {
            parsedImporter = normalizePath(path__default["default"].resolve(inputDir, parseVirtualPagePath(importer)));
        }
        else if (isUniComponentUrl(importer)) {
            parsedImporter = normalizePath(path__default["default"].resolve(inputDir, parseVirtualComponentPath(importer)));
        }
        else if (importer.endsWith('.vue') && !uniCliShared.isMiniProgramPageFile(importer, process.env.UNI_INPUT_DIR)) {
            virtualImporter = virtualComponentPath(importer);
        }
        if (parsedImporter === id && !virtualImporter) {
            continue;
        }
        if (parsedImporter.endsWith('/pages-json-js')) {
            continue;
        }
        let virtual = [];
        let chain = [];
        if (virtualImporter && virtualImporter !== importer && !isUniPageUrl(virtualImporter)) {
            // 直接找虚拟组件的引用者
            virtual = findImporters({
                id: virtualImporter,
                importChain: [],
                getModuleInfo,
                getModuleIds,
                pre: pre.concat(id),
                inputDir,
                iterativeMap,
            });
        }
        else {
            chain = findImporters({
                id: importer,
                importChain: [],
                getModuleInfo,
                getModuleIds,
                pre: pre.concat(id),
                inputDir,
                iterativeMap,
            });
        }
        const allChain = Array.from(new Set([...chain, ...virtual]));
        importerCache.set(parsedImporter, allChain);
        importChain.push(parsedImporter);
        importChain.push(...allChain);
        importChain = Array.from(new Set(importChain));
    }
    importerCache.set(parsedId, importChain);
    return importChain;
}
function getPagesJson(inputDir, options) {
    const json = uniCliShared.parseJson(tComm.readFileSync(path__default["default"].resolve(inputDir, 'pages.json')));
    const parsed = parsePagesJson(json);
    tComm.saveJsonToLogMore(json, './vite-dispatch-script-pages-json-1-raw.json', {
        max: options?.maxLogNumber,
    });
    tComm.saveJsonToLogMore(parsed, './vite-dispatch-script-pages-json-2-parsed.json', {
        max: options?.maxLogNumber,
    });
    return parsed;
}
function parsePagesJson(json) {
    const { pages = [], subPackages = [] } = json;
    const mainPages = pages.map(item => item.path);
    const subPackagesMap = subPackages.reduce((acc, subPackage) => {
        const { root, pages = [] } = subPackage;
        const subPages = pages.map(page => `${root}/${page.path}`);
        subPages.forEach(page => acc[page] = root);
        return acc;
    }, {});
    return {
        mainPages,
        subPackagesMap,
    };
}
function checkSubPackage({ inputDir, mainPages, subPackagesMap, deps, }) {
    const entry = path__default["default"].resolve(inputDir, ENTRY);
    const appVue = path__default["default"].resolve(inputDir, APP_VUE);
    const list = deps || [];
    const subPackageRoots = Object.values(subPackagesMap);
    const res = list.map((item) => {
        // mani.ts 或者 App.vue 使用
        if (item === entry || item === appVue) {
            return MAIN;
        }
        // 被主包页面引用
        const inMapPage = mainPages.find(page => item === `${path__default["default"].resolve(inputDir, page)}.vue`);
        if (inMapPage) {
            return MAIN;
        }
        // 不一定是分包页面，分包组件引用也当作分包使用了
        const subPackageRoot = subPackageRoots.find(root => item.startsWith(`${path__default["default"].resolve(inputDir, root)}/`));
        if (subPackageRoot) {
            return subPackageRoot;
        }
        // 不能默认为主包。被主包页面使用，和被主包其他文件使用有本质区别
        // 被主包页面使用，不能分发
        // 被主包其他文件使用，最后还是可以分发到分包
        // 如 src/local-logic/1 => src/local-logic/2 => src/views/sche/1
        return '';
    }).filter(item => item);
    return Array.from(new Set(res));
}
function staticImportedByEntry(id, getModuleInfo, cache, importStack = []) {
    if (cache.has(id)) {
        return cache.get(id);
    }
    if (importStack.includes(id)) {
        // circular deps!
        cache.set(id, false);
        return false;
    }
    const mod = getModuleInfo(id);
    if (!mod) {
        cache.set(id, false);
        return false;
    }
    if (mod.isEntry) {
        cache.set(id, true);
        return true;
    }
    const someImporterIs = mod.importers.some(importer => staticImportedByEntry(importer, getModuleInfo, cache, importStack.concat(id)));
    cache.set(id, someImporterIs);
    return someImporterIs;
}

function checkImportError({ inputDir, mainPages, subPackagesMap, iterativeMap, }) {
    if (!iterativeMap)
        return [];
    const errorFiles = {};
    Object.keys(iterativeMap).forEach((file) => {
        const value = iterativeMap[file] || [];
        const subPackages = checkSubPackage({
            inputDir,
            mainPages,
            subPackagesMap,
            deps: [
                ...value.importers,
                ...value.dynamic,
            ],
        });
        const curSubPackages = checkSubPackage({
            inputDir,
            mainPages,
            subPackagesMap,
            deps: [file],
        });
        const curSubPackage = curSubPackages[0] || MAIN;
        // 文件属于子包，被主包引用
        const mainError = curSubPackage !== MAIN && !!subPackages.find(item => item === MAIN);
        // 文件属于子包，被其他子包子包
        const subPackageError = curSubPackage !== MAIN
            && !!subPackages.find(item => item !== MAIN && item !== curSubPackage);
        const isError = mainError || subPackageError;
        if (isError) {
            errorFiles[file] = {
                subPackages,
                curSubPackages,
                mainError,
                subPackageError,
            };
        }
    });
    return errorFiles;
}

let logged = false;
const allJson = {};
const iterativeMap = {};
const dispatchedScriptMap = {};
function dispatchScriptVitePlugin(options) {
    return {
        name: 'vite-plugin-dispatch-script',
        enforce: 'post',
        async config() {
            const inputDir = normalizePath(process.env.UNI_INPUT_DIR || '');
            const dispatchDir = options?.dispatchDir ?? `dispatch-script-${tComm.randomString(8)}`;
            const blackList = options?.blackList;
            const whiteList = options?.whiteList;
            const { mainPages, subPackagesMap, } = getPagesJson(inputDir, options);
            return {
                build: {
                    rollupOptions: {
                        output: {
                            manualChunks: createMoveToVendorChunkFn({
                                mainPages,
                                subPackagesMap,
                                dispatchDir,
                                blackList,
                                whiteList,
                            }),
                        },
                    },
                },
            };
        },
        writeBundle() {
            if (logged)
                return;
            logged = true;
            tComm.saveJsonToLogMore(iterativeMap, './vite-dispatch-script-json-1-iterative.json', {
                max: options?.maxLogNumber,
            });
            tComm.saveJsonToLogMore(allJson, './vite-dispatch-script-json-2-flatten.json', {
                max: options?.maxLogNumber,
            });
            tComm.saveJsonToLogMore(dispatchedScriptMap, './vite-dispatch-script-json-3-dispatched.json', {
                max: options?.maxLogNumber,
            });
            const inputDir = normalizePath(process.env.UNI_INPUT_DIR || '');
            const { mainPages, subPackagesMap, } = getPagesJson(inputDir, options);
            const errorFiles = checkImportError({
                inputDir,
                mainPages,
                subPackagesMap,
                iterativeMap,
            });
            tComm.saveJsonToLogMore({
                errorFiles,
                circular: Array.from(circularSet),
            }, './vite-dispatch-script-error-importation-files.json', {
                max: options?.maxLogNumber,
            });
        },
    };
}
function createMoveToVendorChunkFn({ mainPages, subPackagesMap, dispatchDir, blackList, whiteList, }) {
    const cache = new Map();
    const inputDir = normalizePath(process.env.UNI_INPUT_DIR || '');
    console.log('>>> dispatch script inputDir: ', inputDir);
    return (id, { getModuleInfo, getModuleIds }) => {
        const normalizedId = normalizePath(id);
        const filename = normalizedId.split('?')[0];
        let subPackages = [];
        if (!ANALYZE_BLACK_LIST.find(item => id.includes(item)) && !id.endsWith(PAGES_JSON_JS)) {
            const res = findImporters({
                id,
                importChain: [],
                getModuleInfo,
                getModuleIds,
                pre: [],
                inputDir,
                iterativeMap,
            });
            subPackages = checkSubPackage({
                inputDir,
                mainPages,
                subPackagesMap,
                deps: res,
            });
            allJson[id] = {
                importers: res,
                subPackages,
            };
        }
        // 处理资源文件
        if (DEFAULT_ASSETS_RE.test(filename)) {
            return 'common/assets';
        }
        // 处理项目内的js,ts文件
        if (EXTNAME_JS_RE.test(filename)) {
            if (filename.startsWith(inputDir) && !filename.includes('node_modules')) {
                const chunkFileName = removeExt(normalizePath(path__default["default"].relative(inputDir, filename)));
                if (CHUNK_FILE_NAME_BLACK_LIST.includes(chunkFileName)
                    || uniCliShared.hasJsonFile(chunkFileName) // 无同名的page,component
                ) {
                    console.log('>>> dispatch script chunkFileName: ', chunkFileName);
                    // return chunkFileName;
                    return;
                }
                console.log('>>> dispatch script defaultPath: ', id);
                // return;
            }
            if (!subPackages.includes(MAIN)
                && !shared.isInBlackList(id, blackList)
                && !id.includes('?commonjs-module')
                && (subPackages.length === 1 || shared.isInBlackList(id, whiteList))) {
                const dispatchChunkFileName = id.includes('node_modules/')
                    ? removeExt(filename.replace(/^.*node_modules\//, ''))
                        .replace(/\//g, '_')
                    : removeExt(normalizePath(path__default["default"].relative(path__default["default"].resolve(inputDir), filename)))
                        .replace(/\.\.\//g, '_')
                        .replace(/\//g, '_');
                const targetName = `${subPackages[0]}/${dispatchDir}/${dispatchChunkFileName}`;
                dispatchedScriptMap[id] = {
                    subPackage: subPackages[0],
                    targetName,
                };
                return targetName;
            }
            console.log('>>> dispatch script commonVendor: ', id);
            // 非项目内的 js 资源，均打包到 vendor
            return 'common/vendor-1';
        }
        if (isVueJs(normalizedId)
            || (normalizedId.includes('node_modules')
                && !isCSSRequest(normalizedId)
                // 使用原始路径，格式化的可能找不到模块信息 https://github.com/dcloudio/uni-app/issues/3425
                && staticImportedByEntry(id, getModuleInfo, cache))) {
            return 'common/vendor-1';
        }
    };
}

exports.dispatchScriptVitePlugin = dispatchScriptVitePlugin;
